@@ -7,6 +7,10 @@ from pylops import LinearOperator
 from pylops.basicoperators import Diagonal, Identity
 from pylops.optimization.leastsquares import NormalEquationsInversion, \
     RegularizedInversion
+from pylops.optimization.solver import cgls
+from pylops.optimization.eigs import power_iteration
+from pylops.utils.backend import get_array_module, get_module_name, to_numpy
+
 
 try:
     from spgl1 import spgl1
@@ -185,16 +189,19 @@ def _halfthreshold_percentile(x, perc):
     #return _halfthreshold(x, (2. / 3. * thresh) ** (1.5))
     return _halfthreshold(x, (4. / 54 ** (1. / 3.) * thresh) ** 1.5)
 
+
 def _IRLS_data(Op, data, nouter, threshR=False, epsR=1e-10,
                epsI=1e-10, x0=None, tolIRLS=1e-10,
                returnhistory=False, **kwargs_solver):
     r"""Iteratively reweighted least squares with L1 data term
     """
+    ncp = get_array_module(data)
+
     if x0 is not None:
         data = data - Op * x0
     if returnhistory:
-        xinv_hist = np.zeros((nouter + 1, Op.shape[1]))
-        rw_hist = np.zeros((nouter + 1, Op.shape[0]))
+        xinv_hist = ncp.zeros((nouter + 1, int(Op.shape[1])))
+        rw_hist = ncp.zeros((nouter + 1, int(Op.shape[0])))
 
     # first iteration (unweighted least-squares)
     xinv = NormalEquationsInversion(Op, None, data, epsI=epsI,
@@ -207,9 +214,9 @@ def _IRLS_data(Op, data, nouter, threshR=False, epsR=1e-10,
         # other iterations (weighted least-squares)
         xinvold = xinv.copy()
         if threshR:
-            rw = 1. / np.maximum(np.abs(r), epsR)
+            rw = 1. / ncp.maximum(ncp.abs(r), epsR)
         else:
-            rw = 1. / (np.abs(r) + epsR)
+            rw = 1. / (ncp.abs(r) + epsR)
         rw = rw / rw.max()
         R = Diagonal(rw)
         xinv = NormalEquationsInversion(Op, [], data, Weight=R,
@@ -222,7 +229,7 @@ def _IRLS_data(Op, data, nouter, threshR=False, epsR=1e-10,
             rw_hist[iiter] = rw
             xinv_hist[iiter + 1] = xinv
         # check tolerance
-        if np.linalg.norm(xinv - xinvold) < tolIRLS:
+        if ncp.linalg.norm(xinv - xinvold) < tolIRLS:
             nouter = iiter
             break
 
@@ -243,15 +250,23 @@ def _IRLS_model(Op, data, nouter, threshR=False, epsR=1e-10,
                 returnhistory=False, **kwargs_solver):
     r"""Iteratively reweighted least squares with L1 model term
     """
+    ncp = get_array_module(data)
+
     if x0 is not None:
         data = data - Op * x0
     if returnhistory:
-        xinv_hist = np.zeros((nouter + 1, Op.shape[1]))
-        rw_hist = np.zeros((nouter + 1, Op.shape[0]))
+        xinv_hist = ncp.zeros((nouter + 1, int(Op.shape[1])))
+        rw_hist = ncp.zeros((nouter + 1, int(Op.shape[0])))
 
     Iop = Identity(data.size, dtype=data.dtype)
     # first iteration (unweighted least-squares)
-    xinv = Op.H @ lsqr(Op @ Op.H + (epsI**2) * Iop, data, **kwargs_solver)[0]
+    if ncp == np:
+        xinv = Op.H @ \
+               lsqr(Op @ Op.H + (epsI ** 2) * Iop, data, **kwargs_solver)[0]
+    else:
+        xinv = Op.H @ cgls(Op @ Op.H + (epsI ** 2) * Iop, data,
+                           ncp.zeros(int(Op.shape[0]), dtype=Op.dtype),
+                           **kwargs_solver)[0]
     if returnhistory:
         xinv_hist[0] = xinv
     for iiter in range(nouter):
@@ -259,9 +274,15 @@ def _IRLS_model(Op, data, nouter, threshR=False, epsR=1e-10,
         xinvold = xinv.copy()
         rw = np.abs(xinv)
         rw = rw / rw.max()
-        R = Diagonal(rw)
-        xinv = R @ Op.H @ lsqr(Op @ R @ Op.H + epsI**2 * Iop,
-                               data, **kwargs_solver)[0]
+        R = Diagonal(rw, dtype=rw.dtype)
+        if ncp == np:
+            xinv = R @ Op.H @ lsqr(Op @ R @ Op.H + epsI ** 2 * Iop,
+                                   data, **kwargs_solver)[0]
+        else:
+            xinv = R @ Op.H @ cgls(Op @ R @ Op.H + epsI ** 2 * Iop,
+                                   data,
+                                   ncp.zeros(int(Op.shape[0]), dtype=Op.dtype),
+                                   **kwargs_solver)[0]
         # save history
         if returnhistory:
             rw_hist[iiter] = rw
@@ -283,7 +304,6 @@ def _IRLS_model(Op, data, nouter, threshR=False, epsR=1e-10,
         return xinv, nouter
 
 
-
 def IRLS(Op, data, nouter, threshR=False, epsR=1e-10,
          epsI=1e-10, x0=None, tolIRLS=1e-10,
          returnhistory=False, kind='data',
@@ -331,7 +351,9 @@ def IRLS(Op, data, nouter, threshR=False, epsR=1e-10,
     **kwargs_solver
         Arbitrary keyword arguments for
         :py:func:`scipy.sparse.linalg.cg` solver for data IRLS and
-        :py:func:`scipy.sparse.linalg.lsqr` solver for model IRLS
+        :py:func:`scipy.sparse.linalg.lsqr` solver for model IRLS when using
+        numpy data(or :py:func:`pylops.optimization.solver.cg` and
+        :py:func:`pylops.optimization.solver.cgls` when using cupy data)
 
     Returns
     -------
@@ -495,6 +517,8 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
     between the basis is generally properly handled.
 
     """
+    ncp = get_array_module(data)
+
     Op = LinearOperator(Op)
     if show:
         tstart = time.time()
@@ -509,9 +533,9 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
     # find normalization factor for each column
     if normalizecols:
         ncols = Op.shape[1]
-        norms = np.zeros(ncols)
+        norms = ncp.zeros(ncols)
         for icol in range(ncols):
-            unit = np.zeros(ncols, dtype=Op.dtype)
+            unit = ncp.zeros(ncols, dtype=Op.dtype)
             unit[icol] = 1
             norms[icol] = np.linalg.norm(Op.matvec(unit))
     if show:
@@ -523,7 +547,7 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
         x = []
     cols = []
     res = data.copy()
-    cost = np.zeros(niter_outer + 1)
+    cost = ncp.zeros(niter_outer + 1)
     cost[0] = np.linalg.norm(data)
     iiter = 0
     while iiter < niter_outer and cost[iiter] > sigma:
@@ -543,7 +567,7 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
         # update active set
         if imax not in cols:
             addnew = True
-            cols.append(imax)
+            cols.append(int(imax))
         else:
             addnew = False
             imax_in_cols = cols.index(imax)
@@ -551,8 +575,8 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
         # estimate model for current set of columns
         if niter_inner == 0:
             # MP update
-            Opcol = Op.apply_columns([imax, ])
-            res -= Opcol.matvec([cres[imax], ])
+            Opcol = Op.apply_columns([int(imax), ])
+            res -= Opcol.matvec(cres[imax] * ncp.ones(1))
             if addnew:
                 x.append(cres[imax])
             else:
@@ -560,7 +584,12 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
         else:
             # OMP update
             Opcol = Op.apply_columns(cols)
-            x = lsqr(Opcol, data, iter_lim=niter_inner)[0]
+            if ncp == np:
+                x = lsqr(Opcol, data, iter_lim=niter_inner)[0]
+            else:
+                x = cgls(Opcol, data, ncp.zeros(int(Opcol.shape[1]),
+                                                dtype=Opcol.dtype),
+                         niter=niter_inner)[0]
             res = data - Opcol.matvec(x)
         iiter += 1
         cost[iiter] = np.linalg.norm(res)
@@ -568,21 +597,22 @@ def OMP(Op, data, niter_outer=10, niter_inner=40, sigma=1e-4,
             if iiter < 10 or niter_outer - iiter < 10 or iiter % 10 == 0:
                 msg = '%6g        %12.5e' % (iiter + 1, cost[iiter])
                 print(msg)
-    xinv = np.zeros(Op.shape[1], dtype=Op.dtype)
-    xinv[cols] = np.array(x)
+    xinv = ncp.zeros(int(Op.shape[1]), dtype=Op.dtype)
+    xinv[cols] = ncp.array(x)
     if show:
         print('\nIterations = %d        Total time (s) = %.2f'
               % (iiter, time.time() - tstart))
-        print('-----------------------------------------------------------------\n')
+        print(
+            '-----------------------------------------------------------------\n')
     return xinv, iiter, cost
 
 
 def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
          tol=1e-10, monitorres=False, returninfo=False, show=False,
-         threshkind='soft', perc=None, callback=None):
+         threshkind='soft', perc=None, callback=None, decay=None, SOp=None):
     r"""Iterative Shrinkage-Thresholding Algorithm (ISTA).
 
-    Solve an optimization problem with :math:`Lp, \quad p=0, 1/2, 1`
+    Solve an optimization problem with :math:`L_p, \; p=0, 1/2, 1`
     regularization, given the operator ``Op`` and data ``y``. The operator
     can be real or complex, and should ideally be either square :math:`N=M`
     or underdetermined :math:`N<M`.
@@ -624,7 +654,11 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
     callback : :obj:`callable`, optional
         Function with signature (``callback(x)``) to call after each iteration
         where ``x`` is the current model vector
-
+    decay : :obj:`numpy.ndarray`, optional
+        Decay factor to be applied to thresholding during iterations
+    SOp : :obj:`pylops.LinearOperator`, optional
+        Regularization operator (use when solving the analysis problem)
+    
     Returns
     -------
     xinv : :obj:`numpy.ndarray`
@@ -654,14 +688,23 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
 
     Notes
     -----
-    Solves the following optimization problem for the operator
+    Solves the following synthesis problem for the operator
     :math:`\mathbf{Op}` and the data :math:`\mathbf{d}`:
 
     .. math::
         J = ||\mathbf{d} - \mathbf{Op} \mathbf{x}||_2^2 +
             \epsilon ||\mathbf{x}||_p
 
-    using the Iterative Shrinkage-Thresholding Algorithms (ISTA) [1]_, where
+    or the analysis problem:
+
+    .. math::
+        J = ||\mathbf{d} - \mathbf{Op} \mathbf{x}||_2^2 +
+            \epsilon ||\mathbf{SOp}^H\mathbf{x}||_p
+
+    if ``SOp`` is provided. Note that in the first case, ``SOp`` should be
+    assimilated in the modelling operator (i.e., ``Op=GOp * SOp``).
+
+    The Iterative Shrinkage-Thresholding Algorithms (ISTA) [1]_ is used, where
     :math:`p=0, 1, 1/2`. This is a very simple iterative algorithm which
     applies the following step:
 
@@ -669,6 +712,13 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
         \mathbf{x}^{(i+1)} = T_{(\epsilon \alpha /2, p)} (\mathbf{x}^{(i)} +
         \alpha \mathbf{Op}^H (\mathbf{d} - \mathbf{Op} \mathbf{x}^{(i)}))
 
+    or
+
+    .. math::
+        \mathbf{x}^{(i+1)} = \mathbf{SOp}(T_{(\epsilon \alpha /2, p)}
+        (\mathbf{SOp}^H(\mathbf{x}^{(i)} + \alpha \mathbf{Op}^H (\mathbf{d} -
+        \mathbf{Op} \mathbf{x}^{(i)}))))
+
     where :math:`\epsilon \alpha /2` is the threshold and :math:`T_{(\tau, p)}`
     is the thresholding rule. The most common variant of ISTA uses the
     so-called soft-thresholding rule :math:`T(\tau, p=1)`. Alternatively an
@@ -708,6 +758,13 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
     else:
         threshf = _halfthreshold_percentile
 
+    # identify backend to use
+    ncp = get_array_module(data)
+
+    # prepare decay (if not passed)
+    if perc is None and decay is None:
+        decay = ncp.ones(niter)
+
     if show:
         tstart = time.time()
         print('ISTA optimization (%s thresholding)\n'
@@ -723,8 +780,13 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
             Op = LinearOperator(Op, explicit=False)
         # compute largest eigenvalues of Op^H * Op
         Op1 = LinearOperator(Op.H * Op, explicit=False)
-        maxeig = np.abs(Op1.eigs(neigs=1, symmetric=True, niter=eigsiter,
-                                 **dict(tol=eigstol, which='LM')))[0]
+        if get_module_name(ncp) == 'numpy':
+            maxeig = np.abs(Op1.eigs(neigs=1, symmetric=True, niter=eigsiter,
+                                     **dict(tol=eigstol, which='LM'))[0])
+        else:
+            maxeig = np.abs(power_iteration(Op1, niter=eigsiter,
+                                            tol=eigstol, dtype=Op1.dtype,
+                                            backend='cupy')[0])
         alpha = 1./maxeig
 
     # define threshold
@@ -740,7 +802,7 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
         print(head1)
 
     # initialize model and cost function
-    xinv = np.zeros(Op.shape[1], dtype=Op.dtype)
+    xinv = ncp.zeros(int(Op.shape[1]), dtype=Op.dtype)
     if monitorres:
         normresold = np.inf
     if returninfo:
@@ -766,10 +828,14 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
 
         # update inverted model
         xinv_unthesh = xinv + grad
+        if SOp is not None:
+            xinv_unthesh = SOp.rmatvec(xinv_unthesh)
         if perc is None:
-            xinv = threshf(xinv_unthesh, thresh)
+            xinv = threshf(xinv_unthesh, decay[iiter] * thresh)
         else:
             xinv = threshf(xinv_unthesh, 100 - perc)
+        if SOp is not None:
+            xinv = SOp.matvec(xinv)
 
         # model update
         xupdate = np.linalg.norm(xinv - xinvold)
@@ -787,7 +853,8 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
         if show:
             if iiter < 10 or niter - iiter < 10 or iiter % 10 == 0:
                 msg = '%6g  %12.5e  %10.3e   %9.3e  %10.3e' % \
-                      (iiter+1, xinv[0], costdata, costdata+costreg, xupdate)
+                      (iiter+1, to_numpy(xinv[:2])[0], costdata,
+                       costdata + costreg, xupdate)
                 print(msg)
 
         # check tolerance
@@ -812,13 +879,13 @@ def ISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
 
 def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
           tol=1e-10, returninfo=False, show=False, threshkind='soft',
-          perc=None, callback=None):
+          perc=None, callback=None, decay=None, SOp=None):
     r"""Fast Iterative Shrinkage-Thresholding Algorithm (FISTA).
 
-    Solve an optimization problem with :math:`L1` regularization function given
-    the operator ``Op`` and data ``y``. The operator can be real or complex,
-    and should ideally be either square :math:`N=M` or underdetermined
-    :math:`N<M`.
+    Solve an optimization problem with :math:`L_p, \; p=0, 1/2, 1`
+    regularization, given the operator ``Op`` and data ``y``.
+    The operator can be real or complex, and should ideally be either square
+    :math:`N=M` or underdetermined :math:`N<M`.
 
     Parameters
     ----------
@@ -855,6 +922,10 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
     callback : :obj:`callable`, optional
         Function with signature (``callback(x)``) to call after each iteration
         where ``x`` is the current model vector
+    decay : :obj:`numpy.ndarray`, optional
+        Decay factor to be applied to thresholding during iterations
+    SOp : :obj:`pylops.LinearOperator`, optional
+        Regularization operator (use when solving the analysis problem)
 
     Returns
     -------
@@ -883,14 +954,22 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
 
     Notes
     -----
-    Solves the following optimization problem for the operator
+    Solves the following synthesis problem for the operator
     :math:`\mathbf{Op}` and the data :math:`\mathbf{d}`:
 
     .. math::
         J = ||\mathbf{d} - \mathbf{Op} \mathbf{x}||_2^2 +
             \epsilon ||\mathbf{x}||_p
 
-    using the Fast Iterative Shrinkage-Thresholding Algorithm (FISTA) [1]_,
+    or the analysis problem:
+
+    .. math::
+        J = ||\mathbf{d} - \mathbf{Op} \mathbf{x}||_2^2 +
+            \epsilon ||\mathbf{SOp}^H\mathbf{x}||_p
+
+    if ``SOp`` is provided.
+
+    The Fast Iterative Shrinkage-Thresholding Algorithm (FISTA) [1]_ is used,
     where :math:`p=0, 1, 1/2`. This is a modified version of ISTA solver with
     improved convergence properties and limited additional computational cost.
     Similarly to the ISTA solver, the choice of the thresholding algorithm to
@@ -926,6 +1005,13 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
     else:
         threshf = _halfthreshold_percentile
 
+    # identify backend to use
+    ncp = get_array_module(data)
+
+    # prepare decay (if not passed)
+    if perc is None and decay is None:
+        decay = ncp.ones(niter)
+
     if show:
         tstart = time.time()
         print('FISTA optimization (%s thresholding)\n'
@@ -941,9 +1027,14 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
             Op = LinearOperator(Op, explicit=False)
         # compute largest eigenvalues of Op^H * Op
         Op1 = LinearOperator(Op.H * Op, explicit=False)
-        maxeig = np.abs(Op1.eigs(neigs=1, symmetric=True, niter=eigsiter,
-                                 **dict(tol=eigstol, which='LM')))[0]
-        alpha = 1./maxeig
+        if get_module_name(ncp) == 'numpy':
+            maxeig = np.abs(Op1.eigs(neigs=1, symmetric=True, niter=eigsiter,
+                                     **dict(tol=eigstol, which='LM')))[0]
+        else:
+            maxeig = np.abs(power_iteration(Op1, niter=eigsiter,
+                                            tol=eigstol, dtype=Op1.dtype,
+                                            backend='cupy')[0])
+        alpha = 1. / maxeig
 
     # define threshold
     thresh = eps * alpha * 0.5
@@ -958,11 +1049,11 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
         print(head1)
 
     # initialize model and cost function
-    xinv = np.zeros(Op.shape[1], dtype=Op.dtype)
+    xinv = ncp.zeros(int(Op.shape[1]), dtype=Op.dtype)
     zinv = xinv.copy()
     t = 1
     if returninfo:
-        cost = np.zeros(niter+1)
+        cost = np.zeros(niter + 1)
 
     # iterate
     for iiter in range(niter):
@@ -976,10 +1067,14 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
 
         # update inverted model
         xinv_unthesh = zinv + grad
+        if SOp is not None:
+            xinv_unthesh = SOp.rmatvec(xinv_unthesh)
         if perc is None:
-            xinv = threshf(xinv_unthesh, thresh)
+            xinv = threshf(xinv_unthesh, decay[iiter] * thresh)
         else:
             xinv = threshf(xinv_unthesh, 100 - perc)
+        if SOp is not None:
+            xinv = SOp.matvec(xinv)
 
         # update auxiliary coefficients
         told = t
@@ -990,8 +1085,8 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
         xupdate = np.linalg.norm(xinv - xinvold)
 
         if returninfo or show:
-            costdata = 0.5*np.linalg.norm(data - Op.matvec(xinv))**2
-            costreg = eps*np.linalg.norm(xinv, ord=1)
+            costdata = 0.5 * np.linalg.norm(data - Op.matvec(xinv)) ** 2
+            costreg = eps * np.linalg.norm(xinv, ord=1)
         if returninfo:
             cost[iiter] = costdata + costreg
 
@@ -1000,9 +1095,10 @@ def FISTA(Op, data, niter, eps=0.1, alpha=None, eigsiter=None, eigstol=0,
             callback(xinv)
 
         if show:
-            if iiter < 10 or niter-iiter < 10 or iiter % 10 == 0:
+            if iiter < 10 or niter - iiter < 10 or iiter % 10 == 0:
                 msg = '%6g  %12.5e  %10.3e   %9.3e  %10.3e' % \
-                      (iiter+1, xinv[0], costdata, costdata+costreg, xupdate)
+                      (iiter + 1, to_numpy(xinv[:2])[0], costdata,
+                       costdata + costreg, xupdate)
                 print(msg)
 
         # check tolerance
@@ -1224,7 +1320,7 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
     where :math:`\mu` and :math:`\epsilon_{{R}_{L2,i}}` are the damping factors
     used to weight the different L2 regularization terms of the cost function.
 
-    The generalized Split Bergman algorithm is used to solve such cost
+    The generalized Split-Bergman algorithm [1]_ is used to solve such cost
     function: the algorithm is composed of a sequence of unconstrained
     inverse problems and Bregman updates.
 
@@ -1232,7 +1328,7 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
     problem:
 
     .. math::
-        J = \mu/2 ||\textbf{d} - \textbf{Op} \textbf{x} |||_2^2 +
+        J = \mu/2 ||\textbf{d} - \textbf{Op} \textbf{x}||_2^2 +
         \sum_i \epsilon_{{R}_{L2,i}}/2 ||\mathbf{d_{{R}_{L2,i}}} -
         \mathbf{R_{L2,i}} \textbf{x}||_2^2 +
         \sum_i || \textbf{y}_i ||_1 \quad s.t \quad
@@ -1243,9 +1339,9 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
     .. math::
         (\textbf{x}^{k+1}, \textbf{y}_i^{k+1}) =
         \operatorname*{arg\,min}_{\mathbf{x}, \mathbf{y}_i}
-        ||\textbf{d} - \textbf{Op} \textbf{x} |||_2^2 +
+        ||\textbf{d} - \textbf{Op} \textbf{x}||_2^2 +
         \sum_i \epsilon_{{R}_{L2,i}}/2 ||\mathbf{d_{{R}_{L2,i}}} -
-        \mathbf{R_{L2,i}} \textbf{x}||_2^2 + \sum_i || \textbf{y}_i ||_1
+        \mathbf{R_{L2,i}} \textbf{x}||_2^2 + \sum_i || \textbf{y}_i ||_1 +
         \sum_i \epsilon_{{R}_{L1,i}}/2 ||\textbf{y}_i -
         \mathbf{R_{L1,i}} \textbf{x} - \textbf{b}_i^k||_2^2
 
@@ -1262,6 +1358,8 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
        pp. 323-343. 2008.
 
     """
+    ncp = get_array_module(data)
+
     if show:
         tstart = time.time()
         print('Split-Bregman optimization\n'
@@ -1278,7 +1376,7 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
 
     # L1 regularizations
     nregsL1 = len(RegsL1)
-    b = [np.zeros(RegL1.shape[0]) for RegL1 in RegsL1]
+    b = [ncp.zeros(RegL1.shape[0], dtype=Op.dtype) for RegL1 in RegsL1]
     d = b.copy()
 
     # L2 regularizations
@@ -1286,7 +1384,7 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
     if nregsL2 > 0:
         Regs = RegsL2 + RegsL1
         if dataregsL2 is None:
-            dataregsL2 = [np.zeros(Reg.shape[0]) for Reg in RegsL2]
+            dataregsL2 = [ncp.zeros(Reg.shape[0], dtype=Op.dtype) for Reg in RegsL2]
     else:
         Regs = RegsL1
         dataregsL2 = []
@@ -1296,11 +1394,11 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
              range(nregsL2)] + \
             [np.sqrt(epsRL1s[ireg] / 2) / np.sqrt(mu / 2) for ireg in
              range(nregsL1)]
-    xinv = np.zeros_like(np.zeros(Op.shape[1])) if x0 is None else x0
-    xold = np.inf * np.ones_like(np.zeros(Op.shape[1]))
+    xinv = ncp.zeros(Op.shape[1], dtype=Op.dtype) if x0 is None else x0
+    xold = ncp.full(Op.shape[1], ncp.inf, dtype=Op.dtype)
 
     itn_out = 0
-    while np.linalg.norm(xinv - xold) > tol and itn_out < niter_outer:
+    while ncp.linalg.norm(xinv - xold) > tol and itn_out < niter_outer:
         xold = xinv
         for _ in range(niter_inner):
             # Regularized problem
@@ -1320,16 +1418,16 @@ def SplitBregman(Op, RegsL1, data, niter_outer=3, niter_inner=5, RegsL2=None,
         itn_out += 1
 
         if show:
-            costdata = mu/2. * np.linalg.norm(data - Op.matvec(xinv)) ** 2
+            costdata = mu/2. * ncp.linalg.norm(data - Op.matvec(xinv)) ** 2
             costregL2 = 0 if RegsL2 is None else \
-                [epsRL2 * np.linalg.norm(dataregL2 - RegL2.matvec(xinv)) ** 2
+                [epsRL2 * ncp.linalg.norm(dataregL2 - RegL2.matvec(xinv)) ** 2
                  for epsRL2, RegL2, dataregL2 in zip(epsRL2s, RegsL2, dataregsL2)]
-            costregL1 = [np.linalg.norm(RegL1.matvec(xinv), ord=1)
+            costregL1 = [ncp.linalg.norm(RegL1.matvec(xinv), ord=1)
                          for epsRL1, RegL1 in zip(epsRL1s, RegsL1)]
-            cost = costdata + np.sum(np.array(costregL2)) + \
-                   np.sum(np.array(costregL1))
+            cost = costdata + ncp.sum(ncp.array(costregL2)) + \
+                   ncp.sum(ncp.array(costregL1))
             msg = '%6g  %12.5e       %10.3e        %9.3e' % \
-                  (np.abs(itn_out), xinv[0], costdata, cost)
+                  (ncp.abs(itn_out), ncp.real(xinv[0]), costdata, cost)
             print(msg)
 
     if show:
